# recombass/cli.py
import click
from . import maffilter, nr, pre, pmrcal, wtdenoise

@click.command()
@click.option('-m', '--maf', type=float, help='Max major allele frequency (e.g., 0.95)')
@click.option('-n', '--nonredundant', type=float, help='Non-redundant strain clustering threshold (e.g., 0.01)')
@click.option('-cc', '--cold-criterion', type=float, default=-0.4, show_default=True, help='Cold spot threshold')
@click.option('-hc', '--hot-criterion', type=float, default=0.6, show_default=True, help='Hot spot threshold')
@click.option('-i', '--input', required=True, type=click.Path(exists=True), help='Input SNP matrix (TSV)')
@click.option('-o', '--output', required=True, help='Output prefix')
def main(maf, nonredundant, cold_criterion, hot_criterion, input, output):
    """Recombination detection pipeline from SNP matrix."""
    path = input
    intermediate_path = output

    # Step 1: Non-redundant filtering
    if nonredundant is not None:
        path = nr.process_noredundant(path, intermediate_path, percentage=nonredundant)
        intermediate_path = path  # update for next step

    # Step 2: MAF filtering
    if maf is not None:
        path = maffilter.filter_maf(path, maf_rate=maf, output_path=intermediate_path)

    # Step 3: SNP distance & PMR calculation
    pre.process_snp_dists(path, intermediate_path)
    pmrcal.process_snp_data(path, intermediate_path)

    # Step 4: Wavelet denoising and plotting
    wtdenoise.plot_and_save(
        wtdenoise.get_crange,
        wtdenoise.get_hrange,
        wtdenoise.wavelet_transform,
        result_path=intermediate_path,
        hc=hot_criterion,
        lc=cold_criterion
    )

if __name__ == '__main__':
    main()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import polars as pl
import polars.selectors as cs




def pmr_gap(x, y):
    return x * (y - 1) / 2 + x * (y - x) / 2


def pmr_snp(x, _):
    return (x * x - x) / 2

cal_map = {
    "A": pmr_snp,
    "T": pmr_snp,
    "C": pmr_snp,
    "G": pmr_snp,
    "-": pmr_gap
}

def cal(x, y):
    return cal_map[x[0]](x[1], y)

def pmr_us(df, lenth):
    counts = df.value_counts()
    dist = [cal(x, lenth) for x in counts.items()]
    return np.sum(dist)

def R_dist(x, y):
    if y==x.es:
        return np.nan
    return -1 * (x.pmn - x.es) / (y - x.es)

def remove_top_bottom_10_percent(arr):
    arr = np.array(arr)  # 确保是 numpy 数组
    lower = np.percentile(arr, 10)
    upper = np.percentile(arr, 90)
    return arr[(arr > lower) & (arr < upper)]

def cal_recx(values):
    """
    计算 recx 值
    
    Args:
        values: list 或 array 类型的数据
        
    Returns:
        计算得到的 recx 值
    """
    values = np.array(values)  # 确保是 numpy 数组
    
    # 移除 NaN 值
    values = values[~np.isnan(values)]
    
    # 如果没有有效数据，返回 NaN
    if len(values) == 0:
        return np.nan
    
    # 移除前10%和后10%的数据
    values = remove_top_bottom_10_percent(values)
    
    if len(values) > 0:
        x = np.std(values)
        return np.exp(-10.106*x + 3.388)
    else:
        return np.nan
# 提取核心逻辑为函数
def process_snp_data(oripath,snppath):
    df2 = pl.read_csv(f'{snppath}.fa.dist.tr', separator='\t')
    df2 = df2.with_columns(pl.col('strain1').cast(pl.String))
    df2 = df2.with_columns(pl.col('strain2').cast(pl.String))
    var = df2['distance'].std()
    med = df2['distance'].median()
    cut = 0.9 * med

    sns.displot(df2['distance'].to_numpy())
    plt.axvline(x=cut, c='#FFAC00')
    plt.axvline(x=cut * 0.30, c='#FFAC00')
    plt.axvline(x=cut * 0.60, c='#FFAC00')
    plt.axvline(x=cut * 0.90, c='#FFAC00')
    plt.savefig(f"{snppath}.fa.dist.png", dpi=300)

    df = df2.filter(pl.col('distance') > cut)
    df2 = df2.filter(pl.col('distance') < cut)
    udist = df['distance'].mean()

    data = pl.read_csv(f'{oripath}', separator='\t').select(cs.by_dtype(pl.String).cast(pl.Categorical))
    dft = data.to_pandas().T
    lenth = dft.shape[0]
    pmr_list = np.array([pmr_us(dft[x], lenth) for x in dft.columns])
    list4 = dft.columns.to_list()
    dft = 0

    # 处理不同距离范围的数据
    def process_distance_range(df2, cut, lower, upper,batch_size=1000):
        df = df2.filter((lower * cut < pl.col('distance')) & (pl.col('distance') <= upper * cut))
        if df.shape[0] == 0:
            return 0, 0, 0
        mdist = df['distance'].mean() 
        strain_columns = [col for col in data.columns if col != "pos"]
        list_ur = tuple(map(tuple, df.to_numpy()))
        snp_matrix = data.drop("pos").to_numpy(dtype='<U1')
        strain_to_idx = {name: idx for idx, name in enumerate(strain_columns)}
        try:
            index_pairs = [(strain_to_idx[a], strain_to_idx[b]) for a, b in list_ur]
        except KeyError as e:
            raise ValueError(f"Strain name not found in data: {e}")
        results = []
        for start in range(0, len(index_pairs), batch_size):
            batch = index_pairs[start:start+batch_size]
            i_idx = [p[0] for p in batch]
            j_idx = [p[1] for p in batch]
            
            cols_i = snp_matrix[:, i_idx]      
            cols_j = snp_matrix[:, j_idx]      
            
            match = (
                (cols_i == cols_j) |
                (cols_i == "-") |
                (cols_j == "-")
            )
        results.append(match.sum(axis=0))
        l = len(list_ur)
        return results, l, mdist

    p3, l3, mdist3 = process_distance_range(df2, cut, 0.6, 0.9)
    p2, l2, mdist2 = process_distance_range(df2, cut, 0.3, 0.6)
    p1, l1, mdist1 = process_distance_range(df2, cut, 0.01, 0.3)
    df = df2.filter((0.01*cut>=pl.col('distance'))|(pl.col('distance')>0.9*cut))
    list_ur=tuple(map(tuple,df.to_numpy()))

    p = data.select([((data[:,list_ur[t][0]]==data[:,list_ur[t][1]])|(data[:,list_ur[t][0]]=='-')|(data[:,list_ur[t][1]]=='-')).alias(f'{t}') for t in range(len(list_ur))]).sum_horizontal().to_numpy()
    l =len(list_ur) 

    # 合并结果
    p=p+p1+p2+p3
    l=l+l1+l2+l3
    lenur=lenth*(lenth-1)/2-l
    pmrur=pmr_list-p
    pmr_urlist=pmrur/lenur
    p4=p1+p2+p3
    l4=l1+l2+l3
    with open(f'{snppath}.fa.pmrus.txt','w') as f:
        f.write('\n'.join([str(x) for x in pmr_urlist.tolist()]))
    def f(x):
        return (1+mdist1/udist*(x-1))*l1
    def g(x,y):
        if y==x.pmn:
            return np.nan
        return (x.es-x.pmn)/(y-x.pmn)
    es1=f(pmr_urlist)
    dfr = pd.DataFrame(zip(p1, es1), columns=['pmn', 'es'])
    list1 = dfr.apply(R_dist, y=l1, axis=1).values.tolist()
    list1r=dfr.apply(g, y=l1, axis=1).values.tolist()       
    def f3(x): return (1+(mdist1*l1+mdist2*l2+mdist3*l3)/l4/udist*(x-1))*l4
    es4=f3(pmr_urlist)
    dfr = pd.DataFrame(zip(p4, es4), columns=['pmn', 'es'])
    with open(f'{snppath}.fa.pmr.all.ox.txt', 'w') as file:
            file.write('\n'.join([str(x) for x in p4]))
    with open(f'{snppath}.fa.pmr.all.ex.txt', 'w') as file:
            file.write('\n'.join([str(x) for x in es4]))
    with open(f'{snppath}.fa.pmr.30.ox.txt', 'w') as file:
            file.write('\n'.join([str(x) for x in p1]))
    with open(f'{snppath}.fa.pmr.30.ex.txt', 'w') as file:
            file.write('\n'.join([str(x) for x in es1]))
    list5 = dfr.apply(R_dist, y=l4, axis=1).values.tolist()
    list5r=dfr.apply(g, y=l4, axis=1).values.tolist()
    # 保存结果到文件
    def save_results(list_r, list_gr, output_file):
        dfa = pd.DataFrame(zip(list4, list_r), columns=['pos', 'r'])
        dfa = dfa[dfa.r <= 0]
        dfb = pd.DataFrame(zip(list4, list_gr), columns=['pos', 'r'])
        dfb = dfb[dfb.r >= 0]
        res = pd.concat([dfa, dfb]).sort_values('pos')
        all_positions = pd.DataFrame({'pos': list4})
        
        # 合并所有位置与筛选后的结果，用0填充缺失值
        res_complete = all_positions.merge(res, on='pos', how='left')
        res_complete['r'] = res_complete['r'].fillna(0)
        res_complete = res_complete.sort_values('pos')
        
        with open(output_file, 'w') as f:
            f.write('\n'.join([str(x) for x in res_complete['r'].values]))
        return res['r'].values.tolist()
    save_results(list1, list1r, f'{snppath}.fa.pmr.30.2.txt')
    res=save_results(list5, list5r, f'{snppath}.fa.pmr.all.2.txt')
    print(f'{snppath} recx = {cal_recx(res)}')




import pandas as pd
import numpy as np
from sklearn.cluster import AgglomerativeClustering
import polars as pl
import subprocess as sp
def process_noredundant(input_path: str, output_path: str,percentage:float=0.01,n_jobs: int = 20):
    # 生成FASTA格式数据
    df = pl.read_csv(input_path, separator="\t", row_index_name='pos')
    result = df.select(pl.all().str.join())

    fa=[f'>{result[series].name}\n{result[series].item()}\n' for series in result.columns]
    
    with open(f'{output_path}.fa', 'w') as f:
        f.write(''.join(fa[2:]))  # 跳过前两列
    
    # 执行snp-dists计算
    sp.run(f"snp-dists -j {n_jobs} {output_path}.fa > {output_path}.fa.dist", shell=True)
    distance=pd.read_table(f'{output_path}.fa.dist',index_col=0,header=0)
    clustering=AgglomerativeClustering(n_clusters=None,metric='precomputed',linkage='average',distance_threshold=percentage*np.median(distance.values))
    cluster=clustering.fit(distance)
    dft=pd.Series(cluster.labels_,index=df.columns[2:]).drop_duplicates()
    output_path+=f'.nr{int(percentage*100):02d}'
    df[['']+dft.index.to_list()].write_csv(output_path,include_header=True,separator='\t')
    return output_path



import pywt
import matplotlib.pyplot as plt
import pandas as pd
import matplotlib as mpl
def get_range(df, c, comparison):
    datac = df[df['r'] < c if comparison == '<' else df['r'] > c].index.tolist()
    if not datac:
        return []

    result = [[datac[0]]]
    for i in range(1, len(datac)):
        if datac[i - 1] + 1 == datac[i]:
            result[-1].append(datac[i])
        else:
            result.append([datac[i]])

    fin_range = [[df.loc[i[0], 'loc'], df.loc[i[-1], 'loc']] for i in result]
    return fin_range

def get_crange(df, c=-.4):
    return get_range(df, c, '<')

def get_hrange(df, c=0.6):
    return get_range(df, c, '>')


def wavelet_transform(df, rlist,a,hc=0.6,lc=-0.4):
    rec=[float(x) for x in rlist]
    if a == "30":
        wavefunc = 'db4'
        level = 5
        threshold = 0.8
        
    else :
        wavefunc= 'db2'
        level = 8
        threshold = 0.8
    coeffs = pywt.wavedec(rec, wavefunc,mode='periodic', level=level)
    for i in range(1, len(coeffs)):
        coeffs[i] = pywt.threshold(coeffs[i], threshold * max(coeffs[i]))     
    y_denoised = pywt.waverec(coeffs, wavefunc,mode='periodic',)
    if df.shape[0] %2 !=0:
        y_denoised=y_denoised[:-1]
    y_re=pd.DataFrame(y_denoised,columns=['r'],index=range(df.shape[0]))
    y_re['loc']=df.index
    if a == '30':
        coldspot=y_re[y_re['r']>hc]['loc'].tolist()
    else:
        coldspot=y_re[y_re['r']<lc]['loc'].tolist()
    return rec,y_denoised,y_re,coldspot

def plot_and_save(get_crange, get_hrange, wavelet_transform, result_path, widthf=6/2.54,hc=0.6,lc=-0.4):
    # 导入数据集
    params = {
            "font.size": 6,     # 全局字号
            "axes.spines.right":False,  # 坐标系-右侧线
            "axes.spines.top":False,   # 坐标系-上侧线
            "axes.titlesize":10,   # 坐标系-标题-字号
            "axes.labelsize": 8,  # 坐标系-标签-字号
            "legend.loc":'upper right',  # 图例-位置
            "legend.fontsize": 8,  # 图例-字号
            "xtick.labelsize": 6,  # 刻度-标签-字号
            "ytick.labelsize": 6,  # 刻度-标签-字号
            "xtick.direction":'in',   # 刻度-方向
            "ytick.direction":'in',  # 刻度-方向
            'pdf.fonttype':42,



    }
    mpl.rcParams.update(params)
    for a in ['30','all']:
        df=pd.read_table(result_path,index_col=0,header=0)
        with open(f'{result_path}.fa.pmr.{a}.2.txt','r') as f:
            rlist=f.read().rstrip().splitlines()
        rec, y_denoised ,y_re,coldspot = wavelet_transform(df, rlist,a,hc,lc)
        plt.figure().set_size_inches(widthf,widthf*0.618)
        plt.ylim(-1, 1)
        plt.tick_params(which='major',width=0.5,length=3)

        with open(f'{result_path}.fa.pmr.{a}.coldsnp.txt','w')as f:
            f.write('\n'.join([str(x) for x in coldspot]))
        with open(f'{result_path}.{a}.values.txt','w')as f:
            f.write('\n'.join([str(x) for x in y_denoised]))
        plt.scatter(df.index.tolist(), rec, label='original signal',s=0.3,c="none",alpha=0.1,marker='o',edgecolors="grey",linewidths=0.3)
        plt.plot(df.index.tolist(), y_denoised, label='Denoised signal',linewidth=0.5)
        plt.xlabel('Genome Position (MB)')
        plt.ylabel('R value')
        if a=='30':
            hotspot=get_hrange(y_re,hc)
            if hotspot==[]:
                print('no hot region')
                continue
            n=0
            for i in hotspot:
                if n==0:
                    n+=1
                    plt.fill_between([i[0],i[1]],0.95,1,facecolor = '#FF0000' ,label='hot spots')
                    continue
                plt.fill_between([i[0],i[1]],0.95,1,facecolor = '#FF0000', )
            hotspot=y_re[y_re['r']>hc]['loc'].tolist()
            with open(f'{result_path}.fa.pmr.{a}.hotsnp.txt','w')as f:
                f.write('\n'.join([str(x) for x in hotspot]))

        if a=="all":
            n=0
            coldspot=get_crange(y_re,lc)
            if coldspot==[]:
                print('no cold region')
                continue
            for i in coldspot:
                if n==0:
                    n+=1
                    plt.fill_between([i[0],i[1]],-1,-0.95,facecolor = '#0000FF' ,label='cold spots')
                    continue
                plt.fill_between([i[0],i[1]],-1,-0.95,facecolor = '#0000FF', )
        if a=='30':
            plt.axhline(y=hc,c='r',linewidth=0.5)
        else:
            plt.axhline(y=lc,c='r',linewidth=0.5)
        plt.savefig(f'{result_path}.fa.pmr.{a}.2.wt.l.pdf',dpi=300,bbox_inches='tight')


#!/usr/bin/env python
import polars as pl
import subprocess as sp

def process_snp_dists(input_path: str,output_path: str, n_jobs: int = 20):
    """处理SNP距离计算并生成三角矩阵结果
    
    Args:
        input_path: 输入文件路径
        n_jobs: 并行任务数 (默认4)
    """
    # 生成FASTA格式数据
    df = pl.read_csv(input_path, separator="\t", row_index_name='pos')
    result = df.select(pl.all().str.join())
    fa=[f'>{result[series].name}\n{result[series].item()}\n' for series in result.columns]
    
    with open(f'{output_path}.fa', 'w') as f:
        f.write(''.join(fa[2:]))  # 跳过前两列
    
    # 执行snp-dists计算
    sp.run(f"snp-dists -j {n_jobs} {output_path}.fa > {output_path}.fa.dist", shell=True)
    
    # 处理输出结果

    df = pl.read_csv(f"{output_path}.fa.dist", separator="\t", has_header=True)
    row_names = df.columns[1:]
    output_path = f"{output_path}.fa.dist.tr"
    # 构建三角矩阵数据
    data = []
    for i in range(len(row_names)):
        for j in range(i + 1, len(row_names)):
            data.append({
                "strain1": row_names[i],
                "strain2": row_names[j],
                "distance": df[row_names[i]][j]
            })
    
    # 写入输出文件
    pl.DataFrame(data).write_csv(output_path, separator="\t")
    return output_path
